(***********************************************************)
(*  This file is distributed under the terms of the        *)
(*    GNU Lesser General Public License version 3          *)
(*                                                         *)
(*          Copyright 2010: Damien Pous.                   *)
(*                                                         *)
(***********************************************************)

let (>>) f g = fun x -> g (f x)

(* atoms *)
type atom = [`Var of int | `Rav of int]

(* formula constructors (par and bottom are handled as sequents) *)
type 'a constr = [
  | atom
  | `One 
  | `Ten of 'a * 'a
  | `Opl of 'a * 'a
  | `Wit of 'a * 'a
]

(* formulae *)
type t = [t constr | `Seq of seq]

(* sequents *)
and seq = node Seq.t

(* nodes *)
and node = t constr


(* formula constructions *)
let var i: t = `Var i
let rav i: t = `Rav i
let one: t = `One
let bot: t = `Seq Seq.empty
let ten: t -> t -> t = fun a b -> match a,b with
  | `One,b -> b
  | a,`One -> a
  | a,b -> `Ten(a,b)
let par: t -> t -> t = fun a b -> match a,b with 
  | `Seq a, b when Seq.is_empty a -> b
  | a, `Seq b when Seq.is_empty b -> a
  | `Seq a, `Seq b -> `Seq (Seq.append a b)
  | `Seq a, (#constr as b) -> `Seq (Seq.snoc a b)
  | (#constr as a), `Seq b -> `Seq (Seq.cons a b)
  | (#constr as a), (#constr as b) -> `Seq (Seq.cons a (Seq.singleton b))
let wit: t -> t -> t = fun a b -> `Wit(a,b)
let opl: t -> t -> t = fun a b -> `Opl(a,b)

(* sequent constructions *)
let empty: seq = Seq.empty
let singleton: t -> seq = function
  | `Seq l -> l
  | #constr as a -> Seq.singleton a
let append: seq -> seq -> seq = Seq.append
let cons: t -> seq -> seq = fun a l -> Seq.append (singleton a) l
let snoc: seq -> t -> seq = fun l a -> Seq.append l (singleton a)

(* building a sequent out of a list *)
let seq_of_list l = List.fold_right cons l Seq.empty

(* dual of a formula *)
let rec dual: t -> t = function
  | `Var i -> rav i
  | `Rav i -> var i
  | `One -> bot
  | `Seq l -> dual_seq l
  | `Ten(a,b) -> par (dual b) (dual a)
  | `Wit(a,b) -> opl (dual a) (dual b)
  | `Opl(a,b) -> wit (dual a) (dual b)
and dual_seq (l: seq): t = Seq.fold_right (fun a acc -> ten acc (dual (a :> t))) l one

(* convert a variable number to a string *)
let string_of_var i =
  String.make 1 (char_of_int (int_of_char 'a'+i))

(* print an expression *)
let rec print f = function
  | `One -> Printf.fprintf f "1"
  | `Var i -> Printf.fprintf f "%s" (string_of_var i)
  | `Rav i -> Printf.fprintf f "%s'" (string_of_var i)
  | `Seq l -> Printf.fprintf f "%a" print_seq l
  | `Ten (a,b) -> Printf.fprintf f "(%a*%a)" print a print b
  | `Wit (a,b) -> Printf.fprintf f "(%a&%a)" print a print b
  | `Opl (a,b) -> Printf.fprintf f "(%a+%a)" print a print b

(* print a sequent expression *)
and print_seq f (l: seq) = 
  Printf.fprintf f "[%a]" (Common.print_list ";" print) (Seq.to_list l :> t list)

(* size of a sequent *)
let rec xsize acc = function
  | `Ten(a,b) | `Wit(a,b) | `Opl(a,b) -> xsize (xsize acc a) b
  | `Seq l -> xsize_seq acc l
  | _ -> 1+acc
and xsize_seq acc l = Seq.fold_left xsize acc (l:> t Seq.t)
let size = xsize 0
let size_seq = xsize_seq 0

(* equality test *)
let rec equal a b = match a,b with
  | `One     , `One      -> true
  | `Var i   , `Var j
  | `Rav i   , `Rav j    -> i=j
  | `Ten(a,b), `Ten(u,v)
  | `Opl(a,b), `Opl(u,v)
  | `Wit(a,b), `Wit(u,v) -> equal a u && equal b v
  | `Seq h   , `Seq k    -> equal_seq h k
  | _ -> false
and equal_seq h k = Seq.equal (equal :> node -> node -> bool) h k

(* hashing table module for sequents *)
module Table = Hashtbl.Make(
  struct
    let hash_step acc x = (acc lsl 5) + acc + x
    let rec hash acc: t -> int = function
      | `One -> hash_step acc 1
      | `Var i -> hash_step acc (7+(i lsl 1))
      | `Rav i -> hash_step acc (8+(i lsl 1))
      | `Seq l -> hash_seq acc l
      | `Ten (a,b) -> hash (hash (hash_step acc 3) a) b
      | `Wit (a,b) -> hash (hash (hash_step acc 4) a) b
      | `Opl (a,b) -> hash (hash (hash_step acc 5) a) b
    and hash_seq acc l = Seq.fold_left (fun acc a -> hash (hash_step acc 6) (a:>t)) (hash_step acc 2) l
    let hash = hash_seq 5381
    let equal = equal_seq
    type t = seq
  end)


(* pseudo normal form for sequents *)
module N = struct

  (* generic function for maping atoms *)
  let rec map_atoms_node f = function
    | `One -> `One
    | `Ten(a,b) -> let a = map_atoms f a in `Ten(a, map_atoms f b)
    | `Opl(a,b) -> let a = map_atoms f a in `Opl(a, map_atoms f b)
    | `Wit(a,b) -> let a = map_atoms f a in `Wit(a, map_atoms f b)
    | #atom as v -> f v
  and map_atoms f = function
    | `Seq l -> `Seq (map_atoms_seq f l)
    | #constr as a -> (map_atoms_node f a :> t)
  and map_atoms_seq f l = 
    Seq.fold_left (fun q a -> Seq.snoc q (map_atoms_node f a)) Seq.empty l

  let cpl_of_var = function `Var i -> true,i | `Rav i -> false,i
  let var_of_cpl b i = if b then `Var i else `Rav i

  (* renaming a sequent so that variables are contiguous positive
     numbers, appearing in increasing order, and such that `Var i
     always occurs before `Rav i *)
  let rename_seq l =
    let x = ref 0 in
    let t = Hashtbl.create 19 in
    let map v = 
      let b,i = cpl_of_var v in
	try let c,j = Hashtbl.find t i in var_of_cpl (if c then b else not b) j
	with Not_found -> let j = !x in Hashtbl.add t i (b,j); incr x; `Var j
    in map_atoms_seq map l

  (* removing the identity of names in a sequent *)
  (* let anonymise_sequent = map_atoms_node (function `Var i -> `Var 0 | `Rav i -> `Rav 0)  *)

  (* normal form for sequents *)
  let normalise l = 
    match Seq.fold_splittings 
      (fun h k acc -> 
	 let l = rename_seq (Seq.append k h) in
	   (* Seq.force l; *)
	   match acc with
	     | None -> Some l
	     | Some l' -> Some (min l l')
      ) l None
    with
      | Some l -> l
      | None -> assert false

end
let normalise = N.normalise

(* annotated versions *)
module A = struct

  (* formulae *)
  type 'i t = [('i t * 'i) constr | `Seq of 'i seq]
  
  (* sequents *)
  and 'i seq = ('i node * 'i) Seq.t
  
  (* nodes *)
  and 'i node = ('i t * 'i) constr
  
  (* sequent operations *)
  let singleton = function
    | `Seq l,_ -> l
    | #constr, _ as ai -> Seq.singleton ai
  let cons a l = Seq.append (singleton a) l
  let snoc l a = Seq.append l (singleton a)  
  
  (* size of a sequent *)
  let rec xsize acc = function
    | `Ten((a,_),(b,_)) | `Wit((a,_),(b,_)) | `Opl((a,_),(b,_)) -> xsize (xsize acc a) b
    | `Seq l -> xsize_seq acc l
    | _ -> 1+acc
  and xsize_seq acc = Seq.fold_left (fun acc (x,_) -> xsize acc (x:> _ t)) acc
  let size x = xsize 0 x
  let size_seq l = xsize_seq 0 l
  
end

let rec clear_node = function
  | `Var i -> `Var i
  | `Rav i -> `Rav i
  | `One -> `One
  | `Ten((a,_),(b,_)) -> `Ten(clear a, clear b)
  | `Opl((a,_),(b,_)) -> `Opl(clear a, clear b)
  | `Wit((a,_),(b,_)) -> `Wit(clear a, clear b)
and clear: 'i A.t -> t = function
  | `Seq l -> `Seq (clear_seq l)
  | #constr as a -> (clear_node a :> t)
and clear_seq l = 
  Seq.fold_right (fun (a,_) q -> Seq.cons (clear_node a) q) l Seq.empty

