(***********************************************************)
(*  This file is distributed under the terms of the        *)
(*    GNU Lesser General Public License version 3          *)
(*                                                         *)
(*          Copyright 2010: Damien Pous.                   *)
(*                                                         *)
(***********************************************************)

open Formula

(* options to be printed *)
let tmax            = Param.float_option "tmax" "float maximum time to spend on a sequent (s)" 1.0
let cmax            = Param.int_option   "cmax"   "nat maximum number of calls to perform from a sequent" 10000000
let opt             = Param.set_option   "opt"       " use the square optimisation"
let ful             = Param.set_option   "ful"       " do not use the square optimisation (default)" ~p:false

(* other options *)
let variables       = Param.ilist_option "v"      "set number of potential variables for generated sequents" [26]
let size            = Param.ilist_option "s"      "set size of generated sequents (number of leaves)" [30]

(* actions *)
let run_sequents    = Param.set_option ~p:false "rseq" " display (pseudo)random sequents"
let run_corr        = Param.set_option ~p:false "corr" " correlation between spent time and calls number"
let run_proportions = Param.set_option ~p:false "prop" " global proportion of square sequents"
let run_raw_timings = Param.set_option ~p:false "rawt" " raw timing data"
let run_dist        = Param.set_option ~p:false "dist" " time distribution"
let run_disc        = Param.set_option ~p:false "disc" " synchronous calls distribution"
let run_timings     = Param.set_option ~p:false "time" " time statistics"
let run_sized_calls = Param.set_option ~p:false "size" " size-indexed calls statistics"
let run_calls       = Param.set_option ~p:false "call" " synchronous calls statistics"
let run_refl_tests  = Param.set_option ~p:false "refl" " run `reflexive' tests"

let _ = Sys.signal Sys.sigusr1 (Sys.Signal_handle (fun _ -> flush stdout))
let _ = Sys.catch_break true
let _ = Arg.parse (Arg.align !Param.options) (fun s -> raise (Arg.Bad s)) "usage: utas [options]" 
let _ = if !opt && !ful then raise (Arg.Bad "please choose between -ful and -opt")

let sanity_checks() =
  Sanity.check "Typer.square" (Typer.square Sanity.variables) exit 1 Sanity.squares;
  Sanity.check "ful_search" (Search.run (fun _ -> true)) exit 1 Sanity.provables;
  Sanity.check "opt_search" (Search.run (Typer.square Sanity.variables)) exit 1 Sanity.provables;
  Printf.fprintf stderr "sanity check passed\n%!"

let percent a b = 100. *. (float_of_int a) /. (float_of_int b)

let plot f = 
  List.iter (fun v -> 
  List.iter (fun s -> 
	       f v s; flush stdout; 
	       Printf.fprintf stderr "v=%i, s=%i\n%!" v s) !size
	    ) !variables

(* computing random sequents and displaying them with their occurence number *)
let sequents v s =
  let t = Hashtbl.create 19 in
  let incr x = 
    try incr (Hashtbl.find t x) 
    with Not_found -> Hashtbl.add t x (ref 1)
  in
    for i = 1 to !Param.number do
      incr (normalise (Rand.sequent v s))
    done;
    Hashtbl.iter (fun l n -> Printf.printf "%i\t%i\t%i\t%a\n" v s !n print_seq l) t

let do_sequents () =
  Printf.printf "# sequents\n# variables\tleaves\toccurences\tsequent\n";
  plot sequents



let heuristic = if !opt then Typer.square else (fun _ _ -> true)

let search_pure v = Search.run (heuristic v)

(* reflexive tests: check that x |- x is derivable for randomly picked x *)
let refl_tests v s =
  for i = 1 to !Param.number do
    let x = Rand.formula v s in
    let l = seq_of_list [dual x; x] in
      try match Common.timeout !tmax (search_pure v) l with
	| None -> Printf.fprintf stderr "timeout on refl %a\n%!" print x
	| Some false -> Printf.fprintf stderr "proof search was wrong on refl %a\n%!" print x; exit 1
	| _ -> ()
      with Sys.Break -> Printf.fprintf stderr "(%i,%i): %i/%i; skipping %a \n%!" v s i !Param.number print x
  done

let do_refl_tests () = plot refl_tests

(* time statistics *)
let time v s =
  let res = Array.make !Param.number None in
    for i = 0 to !Param.number-1 do
      let x = Rand.sequent v s in
	res.(i) <- Common.timeout !tmax (Common.time (search_pure v) >> fst) x
    done;
    res

(* abstracted: just record mean time and standard deviation *)
let timings v s =
  let t = time v s in
  let t,o = Array.fold_left (fun (t,o) -> function None -> t,1+o | Some x -> x::t,o) ([],0) t in
  let t = Array.of_list t in
  let m,d = Common.mean_stddev t in
    Printf.printf "%i\t%i\t%f\t%f\t%f\n" v s m d (percent o !Param.number)

let do_timings () =
  Printf.printf "# timings\n# variables\tleaves\tmean time\tdeviation\t%%timeouts\n";
  plot timings

(* statistics: print the distribution of image of a function [f] *)
let distribution f v s =
  let step = 5. in
  let contract s = int_of_float (ceil (step *. log (s +. 0.0000001))) in
  let expand i = exp (float_of_int i /. step) in
  let n = Hashtbl.create 19 in
  let min = ref max_int in
  let max = ref min_int in
  let incr i = 
    if i > !max then max := i;
    if i < !min then min := i;
    try incr (Hashtbl.find n i) with Not_found -> Hashtbl.add n i (ref 1) 
  in
  let get i = try !(Hashtbl.find n i) with Not_found -> 0 in
  let k = 100. /. float_of_int !Param.number in
    for i = 0 to !Param.number-1 do
      let x = Rand.formula v s in
      let x = seq_of_list [(* dual x;  *)x] in
	match f v x with
	  | None -> ()
	  | Some t -> incr (contract t)
    done;
    let r = ref 0 in
    for i = !min to !max do
      let ni = get i in
	if ni <> 0 then (
	  r := !r + ni; 
	  Printf.printf "%i\t%i\t%F\t%f\t%f\n" v s (expand i) (float_of_int ni *. k) (float_of_int !r *. k)
	)
    done

let do_time_distribution () =
  let search v x = Common.timeout !tmax (Common.time (search_pure v) >> fst) x in
    Printf.printf "# time distribution\n# variables\tleaves\ttime\tprobability (%%)\n";
    plot (distribution search)

let do_call_distribution () =
  let search v x = 
    let c = ref 0 in
    let h = fun l -> incr c; if !c > !cmax then raise Not_found; heuristic v l in
      try ignore(Search.run h x); Some (float_of_int !c) with Not_found -> None
  in
    Printf.printf "# calls distribution\n# variables\tleaves\tcalls\tprobability (%%)\n";
    plot (distribution search)


(* raw: record all timings *)
let print_float_option f = function
  | None -> Printf.fprintf f "-"
  | Some x -> Printf.fprintf f "%f" x

let raw_timings v s =
  Array.iter 
    (Printf.printf "%i\t%i\t%a\n" v s print_float_option)
    (time v s)
    
let do_raw_timings () =
  Printf.printf "# raw timings\n# variables\tleaves\ttime\n";
  plot raw_timings




let search_calls v c = 
  let count f = fun l -> incr c; f l in
    if !opt then Search.run (count (heuristic v))
    else Search.run (fun _ -> incr c; true)      

(* correlation between time spent and number of calls *)
let correlation v s =
  for i = 1 to !Param.number do
    let l = Rand.sequent v s in
    let c = ref 0 in
      try let t,_ = Common.time (search_calls v c) l in
	Printf.printf "%i\t%i\t%i\t%f\t\n" v s !c t
      with Sys.Break -> Printf.fprintf stderr "skip!\n%!\n"
  done

let do_correlation () =
  Printf.printf "# correlation\n# variables\tleaves\tcalls\ttime\n";
  plot correlation


(* compute the proportions of square sequents *)
let proportions v s = 
  let sqr = ref 0 in
    for i = 1 to !Param.number do
      let l = Rand.sequent v s in
	if Typer.square v l then incr sqr
    done;
    Printf.printf "%i\t%i\t%f\n" v s (percent !sqr !Param.number)

let do_proportions () =
  Printf.printf "# proportions\n# variables\tleaves\t%%square\n";
  plot proportions


(* building a heuristic recorder according to the selected algorithm *)
exception Timeout
let build_recorder v b i =
  let tic = ref 0 in
  let tac () = incr tic; if !tic > !cmax then raise Timeout in
  let c = b 0 in			(* calls *)
  let s = b 0 in			(* skipped calls *)
    (c,s),tic,
    (fun l -> tac(); let incr = i l in incr c; heuristic v l || (incr s; false))

(* proof searching by annotating the sequent first, with a given heuristic recorder *)
let search_rec h l =
  try Some (Search.run h l) with Timeout | Sys.Break -> None


(* global recorder *)
let global_recorder v = 
  build_recorder v ref (fun _ -> incr)

(* global call statistics *)
let calls v s =
  let (ca,sk),tic,h = global_recorder v in
  let o = ref 0 in			(* timeouts *)
  let p = ref 0 in			(* provables *)
    for i = 1 to !Param.number do
      tic := 0;
      let x = Rand.sequent v s in
	match search_rec h x with
	  | None -> incr o;
	  | Some b -> if b then incr p
    done;
    Printf.printf "%i\t%i\t%i\t%f\t%f\t%f\n" v s 
      !ca (percent !sk !ca) (percent !p !ca) (percent !o !ca)

let do_calls () =
  Printf.printf "# calls\n# variables\tleaves\tcalls\t%%skipped\t%%provable\t%%timeouts\n";
  plot calls


(* size indexed recorder: index collected numbers by the size of the corresponding sequent *)
let sized_recorder v s = 
  build_recorder v
    (Array.make (s+1)) 
    (fun l -> let i = size_seq l in fun t -> t.(i) <- t.(i)+1)

(* size indexed call statistics *)
let sized_calls v s =
  let (ca,sk),tic,h = sized_recorder v s in
    for i = 1 to !Param.number do
      tic := 0;
      let x = Rand.sequent v s in
	match search_rec h x with
	  | None -> Printf.fprintf stderr "%i/%i; skipping %a \n%!" i !Param.number print_seq x
	  | _ -> ()
    done;
    for i = 0 to s do
      Printf.printf "%i\t%i\t%i\t%i\t%i\n" v s i ca.(i) sk.(i)
    done

let do_sized_calls () =
  Printf.printf "# sized-calls\n# variables\tstarting leaves\tleaves\tcalls\tskipped\n";
  plot sized_calls




(* entry point *)
let _ = 
  Param.print(); 
  sanity_checks ();
       if !run_refl_tests  then do_refl_tests ()
  else if !run_corr        then do_correlation ()
  else if !run_proportions then do_proportions ()
  else if !run_timings     then do_timings ()
  else if !run_disc        then do_call_distribution ()
  else if !run_dist        then do_time_distribution ()
  else if !run_sequents    then do_sequents ()
  else if !run_raw_timings then do_raw_timings ()
  else if !run_calls       then do_calls ()
  else if !run_sized_calls then do_sized_calls ()
